function SocNetParallelFile(filename, distributedQ, fractionPeople)
% SOCNETPARALLEL Social networks parallel MATLAB program using sparse matrices
% SocNetParallel(filename, distributedQ, fractionPeople)
% is parallel social networks function using sparse matrices and spmd.
% Pre:
%    filename is the base of the file name.  Appended to filename is
%       _nlabs_m.dat, where n is the number of workers and m indicates the
%       file for worker m.
%       Each lab's file has all the activity information for its people, 
%       that is, no person (Person ID) appears in 2 distinct files. 
%    distributedQ is true if distributed programming is to be used and false
%       if local parallel programming is to be used.
%    fractionPeople is fraction of people who should be dominated.
% Post:  The function returns:
%    locations is a minimum dominating set to dominate "fractionPeople".
%        fraction of people, that is, minimal set of locations where that fraction of the people go
%    people is the set of people dominated by "locations".
%    degs is a vector of all degrees of people in a people-to-people connection matrix. 
%    coeffLst is a vector of clustering coefficients of people in a people-to-people connection matrix. 
%
% File SocNetParallel.m
% Author:  Angela B. Shiflet
% 2 April, 2011
% 
% To accompany
% "Getting the 'Edge' on the Next Flu Pandemic:  We Should'a 'Node' Better"
% By Angela B. Shiflet and George W. Shiflet
% Wofford College, Spartanburg, South Carolina
%  2011
%
% Based on
% Eubank, S., V.S. Anil Kumar, M. Marathe, A. Srinivasan and N. Wang.  2004.  
% Structural and Algorithmic Aspects of Large Social Networks. Proc. 15th
% ACM-SIAM Symposium on Discrete Algorithms (SODA), pp. 711-720.
%
% Data downloaded from
% http://ndssl.vbi.vt.edu/opendata/download.php
% 
% NDSSL (Network Dynamics and Simulation Science Laboratory, Virginia Polytechnic Institute 
%      and State University). 2009. "NDSSL Proto-Entities" http://ndssl.vbi.vt.edu/opendata/  
%      Accessed 8/27/9.
% _____. 2009. Synthetic Data Products for Societal Infrastructures and Proto-Populations: Data Set 1.0.
%      ndssl.vbi.vt.edu/Publications/ndssl-tr-06-006.pdf
% _____. 2009. Synthetic Data Products for Societal Infrastructures and Proto-Populations: Data Set 2.0.
%      ndssl.vbi.vt.edu/Publications/ndssl-tr-07-003.pdf 
% _____. 2009. Synthetic Data Products for Societal Infrastructures and Proto-Populations: Data Set 3.0.
%      ndssl.vbi.vt.edu/Publications/ndssl-tr-07-010.pdf

%% Connection matrix
% Form lab's connection matrix, vector of peron IDs, vector of location IDs,
% vector of location degrees

tic;

n = matlabpool('size');  % number of participating labs  

% save setup and timing information
if distributedQ
    dirStr = '\\storage01\matlab\spec101\';
else
    dirStr = '';
end
setupFID = fopen([dirStr 'out_setup_' filename '_' num2str(n) 'Labs.txt'], 'w');
fprintf(setupFID, 'Number of participating worker labs in spmd = %d\n', n);

spmd
% Read file, such as sample_activities_2labs_"labindex".dat, where each line contains
% Household ID, Person ID, Activity ID, Purpose, Start Time, Duration, Location
% OR read a preprocessed file that contains data for the 2nd (Person ID) and 
% 7th (Location) columns
% Read only IDs of people and IDs of locations into an array,
% my_activities, with people IDs on the first column and location IDs on the second
% Assume each lab's file has all the activity information for its people, 
% that is, no person (Person ID) appears in 2 distinct files. 
    
    my_filename = [filename '_' num2str(labindex) '.dat'];
    fid = fopen(my_filename);
    if fid < 0
        disp(['Cannot find file ' my_filename]);
    end
    
    % read my file
    my_numActivities = 0;
    while ~feof(fid)
        % Use the following command when reading an original 7-column activities file
%        currRow = fscanf(fid, '%*d %d %*d %*d %*d %*d %d', [1 2]);
        % Use the following command when reading a 2-column activities file
        currRow = fscanf(fid, '%d %d', [1 2]);
        if ~isempty(currRow)
           my_numActivities = my_numActivities + 1;
           my_activities(my_numActivities, :) = currRow;
        end
    end
    fclose(fid);

    my_personIdLst = unique(my_activities(:, 1))';   % row vector of my people 
    my_numPersons = length(my_personIdLst);
    my_locationIDLst = unique(my_activities(:, 2))'; % row vector of my locations
    my_numLocations = length(my_locationIDLst);
    
    % build my local connection matrix as a sparse matrix
    my_connMat = sparse(zeros(my_numPersons, my_numLocations));
    for i = 1:my_numActivities
        personID = my_activities(i, 1);
        locationID = my_activities(i, 2);
        row = personID == my_personIdLst;
        column = locationID == my_locationIDLst;
        my_connMat(row, column) = 1;
    end;

    %  Get a list of location degrees for my connection matrix
    %  Note:  these are not necessarily the global location degrees
    %  If a person, stored on another lab, goes to this location,
    %  the local degree of the location will be less than the global
    %  degree.
    my_locationDegrees = zeros(1, my_numLocations);
    for j = 1:my_numLocations
        my_locationDegrees(j) = nnz(my_connMat(:, j));
    end;
    
    my_locationLab = labindex*ones(1, my_numLocations);
end;

% clear arrays no longer needed
for i = 1:n
    clear(['my_activities{' num2str(i) '}']);
end;

% save setup information
fprintf(setupFID, 'Input file for labindex 1 = %s\n', my_filename{1});
for i = 1:n
    fprintf(setupFID, 'Lab %d my_numActivities = %d\n', i, my_numActivities{i});
    fprintf(setupFID, 'Lab %d my_numPersons = %d\n', i, my_numPersons{i});
    fprintf(setupFID, 'Lab  %d my_numLocations = %d\n', i, my_numLocations{i});
end;

%%
% Client determines size of pool, gathers global vector of location IDs
% with repetition, global vector of local location degrees, global vector of
% corresponding lab numbers
numPersons = sum([my_numPersons{1:n}]);
fprintf(setupFID, 'Client, numPersons = %d\n', numPersons);

locationDegrees = [my_locationDegrees{1:n}];
% clear arrays no longer needed
for i = 1:n
    clear(['my_locationDegrees{' num2str(i) '}']);
end;

locationLab = [my_locationLab{1:n}];
% clear arrays no longer needed
for i = 1:n
    clear(['my_locationLab{' num2str(i) '}']);
end;

locationIDLst = [my_locationIDLst{1:n}];
% save setup information
fprintf(setupFID, 'Client, size of locationIDLst = %d\n', size(locationIDLst, 2));

%% Minimum dominating set
%
% Client determines vector of locations that appear on more than 1 lab
overlaps = [];
for i = 1:n
    for j = i+1:n
        overlaps = [overlaps intersect(my_locationIDLst{i}, my_locationIDLst{j})];
    end
end
overlaps = unique(overlaps);

% Client updates vector of location degrees to be accurate for locations 
% on more than one lab
for (loc = overlaps)
    whereLoc = find(locationIDLst == loc);
    locationDegrees(whereLoc) = sum(locationDegrees(whereLoc));
end

% Client forms triples of location, location degree, lab
% sorted in descending order by location degree and then sorted
% in ascending order by location ID
sortedLocDegLabLst = sortrows( [locationIDLst; locationDegrees; locationLab]', [-2 1]);

% Get minimum dominating set and locations they dominate using FastGreedy Algorithm
locations = [];  % minimum dominating set to dominate "fractionPeople" fraction of people
                      % that is, minimal set of locations where that fraction of the people go
people = [];     % set of people dominated by "locations"
locDegLabIndex = 1;  % index into sortedLocDegLabLst
fractionLength = fractionPeople * numPersons;
while(length(people) < fractionLength)
    loc = sortedLocDegLabLst(locDegLabIndex, 1);
    locations = [locations loc];
    % find end of sublist of rows with location loc
    endIndex = locDegLabIndex + 1;
    while (loc == sortedLocDegLabLst(endIndex, 1))
        endIndex = endIndex + 1;
    end;
    endIndex = endIndex - 1;

    spmd
        % get the people in my lab who went to this location.
        locationIndex = loc == my_locationIDLst;
        peopleIndices = my_connMat(:, locationIndex) == 1;
        my_people = my_personIdLst(peopleIndices);
    end
    
    % Client gathers people from labs, eliminating duplicates
    people = unique([people [my_people{1:n}]]);
    locDegLabIndex = endIndex + 1;
end;

% clear arrays no longer needed
clear('sortedLocDegLabLst');
for i = 1:n
    clear(['my_locationIDLst{' num2str(i) '}']);
    clear(['my_personIdLst{' num2str(i) '}']);
    clear(['my_people{' num2str(i) '}']);
end;

% save people and locations in files
peopleFID = fopen([dirStr 'out_people.txt'], 'w');
fprintf(peopleFID, '%d ', people);
fclose(peopleFID);

clear('people');

locationsFID = fopen([dirStr 'out_locations.txt'], 'w');
fprintf(locationsFID, '%d ', locations);
fclose(locationsFID);

clear('locations');

%% People-to-people graph
% 
% Each lab stores the complete columns for its people
starts = [cumsum([1, [my_numPersons{1:n - 1}]]), numPersons + 1];
spmd
    my_connPeopleMat = sparse(zeros(numPersons, my_numPersons));
    my_start = starts(labindex);  % starting index of my people in globally considered
                                              % people-to-people connection matrix
    my_end = my_start + my_numPersons - 1;  % ending index of my people
    % account for locations on this lab only
    oringCol = sparse(zeros(numPersons, 1));
    for locIndex = 1:size(my_connMat, 2)
        oringCol(my_start:my_end, 1) = my_connMat(:, locIndex);
        col = find(my_connMat(:, locIndex) == 1);
        for i = col'
            my_connPeopleMat(:, i) = or(oringCol, my_connPeopleMat(:, i));
        end
    end
end

%% account for locations on more than one lab

% For columns in my_connMat, where location are on more than 1 lab
for (loc = overlaps)
    % have each lab find its part to OR
    spmd
        my_oringCol = sparse(zeros(numPersons, 1));
        my_whereLoc = find(my_locationIDLst == loc);
        if my_whereLoc
            my_oringCol(my_start:my_end, 1) = ...
                or(my_oringCol(my_start:my_end, 1), ...
                my_connMat(:, my_whereLoc));
            my_col = find(my_connMat(:, my_whereLoc) == 1);
        end
    end
       
    % construct the column to OR
    oringCol = sparse(zeros(numPersons, 1));
    for i = 1:n
        oringCol = or(oringCol, my_oringCol{i});
    end
    
    % OR that column into all the columns that need it
    spmd    
        if my_whereLoc
            for i = my_col'
               my_connPeopleMat(:, i) = or(oringCol, my_connPeopleMat(:, i));
            end
        end      
    end
    
    % zero out diagonal elements
    spmd
        for i = 1:my_numPersons
            my_connPeopleMat(my_start + i - 1, i) = 0;
        end
    end
end

% clear arrays no longer needed
clear('overlaps');
clear('oringCol');
for i = 1:n
    clear(['my_col{' num2str(i) '}']);
    clear(['my_oringCol{' num2str(i) '}']);
    clear(['my_whereLoc{' num2str(i) '}']);
    clear(['my_connMat{' num2str(i) '}']);
end;

%% degree distribution of people-to-people graph

spmd
    % Obtain vector of degrees of all my people
    % The degree of person ID is the sum of the elements in its column.
    my_degs = sum(my_connPeopleMat(:, 1:my_numPersons));
end
degs = [my_degs{1:n}];  % global vector of degrees

% clear arrays no longer needed
for i = 1:n
    clear(['my_degs{' num2str(i) '}']);
end;

% save degs in file
degsFID = fopen([dirStr 'out_degs.txt'], 'w');
fprintf(degsFID, '%d ', full(degs));
fclose(degsFID);

%% Clustering coefficients - adjacent nodes and denominators

% Clustering coefficient of a vertex is a measure of how close the vertex and 
% its adjacent nodes are to forming a complete graph.
% Suppose adj(i) is the set of nodes adjacent to i.
% The clustering coefficient, C(i), for node i is defined to be the number of edges 
% (u, w), where u, w is in adj(i), divided by the number of combinations  
% C(deg(i), 2 ) =(deg(i))!/(2! (deg(i)-2)!) = (deg(i))(deg(i) - 1)/2, the maximum 
% number of such edges.
% If a node has degree <=1, then its clustering coefficient is 0.

% Store adjacent nodes of each person in "adj".
% Store denominators for calculation of clustering coefficients in
% "denominators".  If deg(i) < 2, the denominator is 1.
% Otherwise, the denominator is (deg(i))(deg(i) - 1)/2

% save coeffLst in file
coeffLstFID = fopen([dirStr 'out_coeffLst.txt'], 'w');

spmd
    my_range = my_start:my_end;  % index range of my people
    my_adj = cell(1, my_numPersons);  % cell array, where i-th element is a vector of indices 
                                                        % of those adjacent
                                                        % to my corresponding person i
    for i = 1:my_numPersons
        my_adj{i} = find(my_connPeopleMat(:, i));  % where are the ones
    end
end

adj = [my_adj{1:n}];

% clear arrays no longer needed
for i = 1:n
    clear(['my_adj{' num2str(i) '}']);
end;

%% Complete clustering coefficients - calculate numerators and quotients

% For each person, i, the numerator is the number of edges (u, w), 
% where u, w is in adj(i).
% Each lab calculates its part.
for i = 1:numPersons
   coeff = 0;
   if (degs(i) > 1)
        spmd
            my_part = intersect(my_range, adj{i});
            if (~isempty(my_part))
                my_partAdjusted = my_part - my_start + 1;
                my_partial = nnz( my_connPeopleMat(adj{i}, my_partAdjusted) );
            else
                my_partial = 0;
            end
         end
        coeff = (sum([my_partial{1:n}])) / (degs(i)*(degs(i) - 1));
    end
    fprintf(coeffLstFID, '%.2f ', coeff);
end

fclose(coeffLstFID);

endTime = toc;

% save setup and timing information
fprintf(setupFID, 'Time to execute SocNetParallel = %f\n', endTime);
fclose(setupFID);

end